class FamilyMember:
	"""
	A FamilyMember Class. Stores the information for a specific member of the family.

	family_member_id: within Familt unique id (String)
	family_id: id for the Family the FamilyMember is within (String)
	sex: Sex (1 = Male, 2 = Female) (Integer)
	affected: Whether the family member is affected (Boolean)
	mum: Another FamilyMember instance which is the FamilyMember's mother.
	dad:  Another FamilyMember instance which is the FamilyMember's dad.
	proband: Whether the FamilyMember is the FamilyMember of interest in this family.

	"""
	
	def __init__(self, family_member_id, family_id, sex, affected, mum=None, dad=None, proband=False):
		
		self.family_member_id = family_member_id
		self.family_id = family_id
		self.sex = sex
		self.affected = affected
		self.mum = mum
		self.dad = dad
		self.proband = proband


		self.check_valid()
		
	def __repr__(self):
		 return self.family_member_id
		
	def get_id(self):
		"""
		Return the FamilyMember family_member_id

		Returns:

			family_member_id: family_member_id
		"""
		
		return self.family_member_id
	
	def get_mum(self):
		"""
		Return the FamilyMember mum attribute

		Returns:

			mum: mum
		"""
		
		return self.mum
	
	def get_dad(self):
		"""
		Return the FamilyMember dad attribute

		Returns:

			dad: dad
		"""
		
		return self.dad

	def check_valid(self):
		"""
		Checks the constructor inputs are valid.
		"""

		if self.sex not in [1,2]:

			raise ValueError(f'The sex of the family member must be either 1 (Male) or 2 (Female). The {self.family_id} object was {self.sex}')

		if self.affected not in [True, False]:

			raise ValueError(f'The affected status of the family member must be either True or False. The {self.family_id} object was {self.affected}')

		if isinstance(self.family_member_id, str) == False:

			raise ValueError(f'The family_member_id attribute must be a string.')

		if isinstance(self.family_id, str) == False:

			raise ValueError(f'The family_id attribute must be a string.')

		if self.mum != None:

			if isinstance(self.mum, FamilyMember) == False:

				raise ValueError(f'The mother must be a valid FamilyMember instance. You entered {self.mum}')

			if self.mum.sex != 2:

				raise ValueError(f'The mother must be female (2). You entered {self.mum.sex}')

			if self.mum.family_id != self.family_id:

				raise ValueError(f'The mother must have the same family_id as this object. The mother\'s family_id was {self.mum.family_id}')

			if self.mum.get_id() == self.get_id():

				raise ValueError(f'FamilyMember cannot be own mother.')

		if self.dad != None:

			if isinstance(self.dad, FamilyMember) == False:

				raise ValueError(f'The father must be a valid FamilyMember instance. You entered {self.dad}')

			if self.dad.sex != 1:

				raise ValueError(f'The mother must be female (2). You entered {self.dad.sex}')

			if self.dad.family_id != self.family_id:

				raise ValueError(f'The father must have the same family_id as this object. The father\'s family_id was {self.dad.family_id}')

			if self.dad.get_id() == self.get_id():

				raise ValueError(f'FamilyMember cannot be own father.')

		if self.proband == True and self.affected == False:

			raise ValueError(f'FamilyMember object cannot be the proband and be unaffected.')









from pyvariantfilter.variant import Variant
from pyvariantfilter.family import Family
from pyvariantfilter.utils import parse_csq_field, get_info_field_dict, compound_het_pair_pass_filter
from pysam import VariantFile
import itertools
import pandas as pd

class VariantSet:
	"""
	A VariantSet object allows multiple variants within a single family to be associated with each other.
	"""
	
	def __init__(self):
		
		self.candidate_compound_het_dict = {}
		self.filtered_compound_het_dict = {}
		self.variant_dict = {}
		self.final_compound_hets = {}
		self.family = None

	def add_family(self, family):
		"""
		Add a family to a VariantSet

		Input:

			family: (Family) - A family object to associate with the VariantSet

		Returns:

			None 

		"""

		assert isinstance(family, Family)
		assert bool(self.variant_dict) == False
		self.family = family
	
	def add_variant(self, variant):
		"""
		Add a variant to a VariantSet.

		Input:
		
			variant: (Variant) - A Variant object to add to the variant set.

		"""

		assert variant.family.family_id == self.family.family_id
		
		if variant.variant_id not in self.variant_dict:
		
			self.variant_dict[variant.variant_id] = variant


	def read_variants_from_vcf(self, vcf_file, parse_csq=True, vep_csq_key='CSQ', proband_variants_only=True, filter_func=None, args=None):
		"""
		Read variants from a standard VCF. Must have AD,GQ and DP fields in the Format section for each sample.

		Input:

			vcf_file: (String) Path to the VCF file to read.
			parse_csq: (Boolean) Whether to attempt to parse the CSQ field added by VEP.
			vep_csq_key: (String) The key of the CSQ field in the VCF INFO section.
			proband_variants_only (Boolean) Only load variants which the proband has an alt allele.
			import_filtered (Boolean) Whether to import variants which fail the VCF Filter

		Returns:

			None - loads variants into self.variant_dict

		"""

		valid_chroms = {'1': None, '2': None, '3': None, '4': None,
		 '5': None, '6': None, '7': None, '8': None, '9': None,
		 '10': None, '11': None, '12': None, '13': None,
		 '14': None, '15': None, '16': None, '17': None,
		 '18': None, '19': None, '20': None, '21': None,
		 '22': None, 'X': None, 'Y': None, 'MT': None, 'M': None}

		assert self.family != None

		family_member_ids = self.family.get_all_family_member_ids()

		if proband_variants_only == True:

			proband_id = self.family.get_proband().get_id()

		bcf_in = VariantFile(vcf_file)

		if parse_csq == True:

			csq_fields = str(bcf_in.header.info[vep_csq_key].record)

			csq_fields = csq_fields = csq_fields.strip()

			index = csq_fields.index('Format:') + 8

			csq_fields = csq_fields[index:len(csq_fields)-2].split('|')

		for rec in bcf_in.fetch():
			
			chrom = rec.chrom
			pos = rec.pos
			ref = rec.ref
			alt = rec.alts
			filter_status = rec.filter.keys()
			info = rec.info
			quality = rec.qual

			if 'chr' in chrom:

				chrom = chrom.strip('chr')

			# Chromosome is correct
			if chrom not in valid_chroms:

				print (f'{chrom} is not a valid chromosome. Not entered into variant set.')

				continue

			info_dict = get_info_field_dict(info, vep_csq_key)

			assert len(alt) == 1

			alt = alt[0]

			if alt == '*':
				continue

			if parse_csq == True:

				csq = rec.info[vep_csq_key]

				transcript_annotations = parse_csq_field(csq, csq_fields)

			new_variant = Variant(chrom=chrom, pos=pos, ref=ref, alt=alt, filter_status=filter_status, quality=quality)
			new_variant.add_family(self.family)
			new_variant.add_transcript_annotations(transcript_annotations)
			new_variant.add_info_annotations(info_dict)

			for family_member_id in family_member_ids:

				sample_genotype_data = rec.samples[family_member_id]
				
				ref_and_alt = [ref, alt]
				
				gts =[]
				ads =[]
				
				for allele in sample_genotype_data['GT']:
					
					if allele == None:
						
						gts.append('.')
						
					else:
						
						gts.append(ref_and_alt[allele])

				if gts[0] == '.' and gts[1] == '.':

						ads.append(0)
						ads.append(0)

				elif len(sample_genotype_data['AD']) == 1 and sample_genotype_data['AD'][0] == None:

						ads.append(0)
						ads.append(0)

				else:

					assert len(sample_genotype_data['AD']) ==2

					for ad in sample_genotype_data['AD']:
					
						if ad == None:
						
							ads.append(0)
						
						else:
						
							ads.append(ad)
				try:

					gq = sample_genotype_data['GQ']

				except:

					print (f'Warning No GQ for variant {new_variant.variant_id}. This could cause filtering errors.')
					# set to high so we don't accidently filter out
					gq = 100

				if gq == None:

					gq = 0

				dp = sample_genotype_data['DP']

				if dp == None:

					dp  = 0
				
				new_variant.add_genotype(family_member_id, gts, ads, gq, dp)

			passes_filter = True

			if filter_func != None and args != None:

				passes_filter = filter_func(new_variant, *args)

				assert passes_filter == True or passes_filter == False

			if proband_variants_only == True:

				if new_variant.has_alt(proband_id) == True and passes_filter == True:

					self.add_variant(new_variant)

			elif passes_filter == True:

				self.add_variant(new_variant)

	def read_variants_from_platypus_vcf(self, vcf_file, parse_csq=True, vep_csq_key='CSQ', proband_variants_only=True, filter_func=None, args=None):
		"""
		Read variants from a platypus VCF. Must have NR, NV and GQ fields in the Format section for each sample.

		Input:

			vcf_file: (String) Path to the VCF file to read.
			parse_csq: (Boolean) Whether to attempt to parse the CSQ field added by VEP.
			vep_csq_key: (String) The key of the CSQ field in the VCF INFO section.
			proband_variants_only (Boolean) Only load variants which the proband has an alt allele.
			import_filtered (Boolean) Whether to import variants which fail the VCF Filter

		Returns:

			None - loads variants into self.variant_dict

		"""
		valid_chroms = {'1': None, '2': None, '3': None, '4': None,
		 '5': None, '6': None, '7': None, '8': None, '9': None,
		 '10': None, '11': None, '12': None, '13': None,
		 '14': None, '15': None, '16': None, '17': None,
		 '18': None, '19': None, '20': None, '21': None,
		 '22': None, 'X': None, 'Y': None, 'MT': None, 'M': None}

		assert self.family != None

		family_member_ids = self.family.get_all_family_member_ids()

		if proband_variants_only == True:

			proband_id = self.family.get_proband().get_id()

		bcf_in = VariantFile(vcf_file)

		if parse_csq == True:

			csq_fields = str(bcf_in.header.info[vep_csq_key].record)

			csq_fields = csq_fields = csq_fields.strip()

			index = csq_fields.index('Format:') + 8

			csq_fields = csq_fields[index:len(csq_fields)-2].split('|')

		for rec in bcf_in.fetch():
			
			chrom = rec.chrom
			pos = rec.pos
			ref = rec.ref
			alt = rec.alts
			filter_status = rec.filter.keys()
			info = rec.info
			quality = rec.qual

			info_dict = get_info_field_dict(info, vep_csq_key)

			if 'chr' in chrom:

				chrom = chrom.strip('chr')

			# Chromosome is correct
			if chrom not in valid_chroms:

				print (f'{chrom} is not a valid chromosome. Not entered into variant set.')

				continue
			
			assert len(alt) == 1

			alt = alt[0]

			if alt == '*':
				continue

			if parse_csq == True:

				csq = rec.info[vep_csq_key]

				transcript_annotations = parse_csq_field(csq, csq_fields)

			new_variant = Variant(chrom=chrom, pos=pos, ref=ref, alt=alt, filter_status=filter_status, quality=quality)
			new_variant.add_family(self.family)
			new_variant.add_transcript_annotations(transcript_annotations)
			new_variant.add_info_annotations(info_dict)


			for family_member_id in family_member_ids:

				sample_genotype_data = rec.samples[family_member_id]
				
				ref_and_alt = [ref, alt]
				
				gts =[]
				ads =[]
				
				for allele in sample_genotype_data['GT']:
					
					if allele == None:
						
						gts.append('.')
						
					else:
						
						gts.append(ref_and_alt[allele])

				if gts[0] == '.' and gts[1] == '.':

						ads.append(0)
						ads.append(0)


				else:

					total_depth = sample_genotype_data['NR']
					variant_depth = sample_genotype_data['NV']

					assert len(total_depth) == 1
					assert len(variant_depth) == 1

					allele_depth_ref = total_depth[0] - variant_depth[0]
					allele_depth_alt = variant_depth[0]


					ads.append(allele_depth_ref)
					ads.append(allele_depth_alt)

				gq = sample_genotype_data['GQ'][0]

				if gq == None:

					gq = 0

				dp = sample_genotype_data['NR'][0]

				if dp == None:

					dp  = 0

				
				new_variant.add_genotype(family_member_id, gts, ads, gq, dp)

			passes_filter = True

			if filter_func != None and args != None:

				passes_filter = filter_func(new_variant, *args)

				assert passes_filter == True or passes_filter == False


			if proband_variants_only == True:

				if new_variant.has_alt(proband_id) == True and passes_filter == True:

					self.add_variant(new_variant)

			elif passes_filter == True:

				self.add_variant(new_variant)
	
				
	def get_candidate_compound_hets(self, feature_key='Feature', consequences={'transcript_ablation': None,
													'splice_acceptor_variant': None,
													'splice_donor_variant': None,
													'stop_gained': None,
													'frameshift_variant': None,
													'stop_lost': None,
													'start_lost': None,
													'transcript_amplification': None,
													'inframe_insertion': None,
													'inframe_deletion': None,
													'missense_variant': None,
													'protein_altering_variant': None,
													'splice_region_variant': None,
													'incomplete_terminal_codon_variant': None,
													'start_retained_variant': None,
													'stop_retained_variant': None,
													'synonymous_variant': None,
													'coding_sequence_variant': None,
													'mature_miRNA_variant': None,
													'5_prime_UTR_variant': None,
													'3_prime_UTR_variant': None,
													'non_coding_transcript_exon_variant': None,
													'intron_variant': None,
													'NMD_transcript_variant': None,
													'non_coding_transcript_variant': None,
													'upstream_gene_variant': None,
													'downstream_gene_variant': None,
													'TFBS_ablation': None,
													'TFBS_amplification': None,
													'TF_binding_site_variant': None,
													'regulatory_region_ablation': None,
													'regulatory_region_amplification': None,
													'feature_elongation': None,
													'regulatory_region_variant': None,
													'feature_truncation': None,
													} ):

		"""
		Go though each variant in the self.variant_dict and build a dictionary of with each transcript as a key 
		and variants within that transcript as a value.

		Only include variants where the proband is heterozozygous, the variant is on an autosome or the X in a female,
		the worst consequence of the variant is in the consequences dict and the variant is not hom_alt in any unaffected 
		samples.

		Input:

			feature_key: (String) The feature to group compound hets on. WARNING - must be unique to a variant - don't use SYMBOL use transcript ID.
			consequences: (Dict) A dictionary containing the consequences to include as a compound het.

		Returns:

			None - loads self.candidate_compound_het_dict

		"""


		assert isinstance(self.family, Family)

		self.candidate_compound_het_dict = {}

		proband = self.family.get_proband()
				
		for variant in self.variant_dict:
			
			if (self.variant_dict[variant].is_on_autosome_or_xfemale() and
				self.variant_dict[variant].is_het(proband.get_id()) and 
				self.variant_dict[variant].get_worst_consequence() in consequences):

				if self.variant_dict[variant].is_hom_alt_in_unaffected() == False:
			
					for gene in self.variant_dict[variant].get_genes(feature_key=feature_key):

						if gene not in self.candidate_compound_het_dict:

							self.candidate_compound_het_dict[gene] = [self.variant_dict[variant]]

						else:

							self.candidate_compound_het_dict[gene].append(self.variant_dict[variant])


	def get_unfiltered_compound_hets_as_dict(self):
		"""
		Flatten the candidate compound het dict. When a gene has two or more variants as values then 
		each of these variants are added to the dict.

		Use for singlton analysis.

		Input: Self

		Returns: None - Loads self.final_compound_hets

		"""

		assert self.candidate_compound_het_dict != False

		self.final_compound_hets = {}

		for gene in self.candidate_compound_het_dict:

			if len(self.candidate_compound_het_dict[gene]) > 1:

				for variant in self.candidate_compound_het_dict[gene]:

					self.final_compound_hets[variant.variant_id] = None

	
	def filter_compound_hets(self, include_both_parents_missing=True,
							 include_denovo=True,
							 allow_hets_in_unaffected=False,
							 check_affected=True):

		"""
		Filter compound hets by phasing them by descent using parental information.

		Input:
		
		include_both_parents_missing (Boolean)
		include_denovo (Boolean)
		allow_hets_in_unaffected (Boolean)
		check_affected (Boolean)

		Returns: None - creates self.filtered_compound_het_dict

		"""

		self.filtered_compound_het_dict = {}
	
		proband = self.family.get_proband()
		proband_id = proband.get_id()
		
		proband_dad = proband.get_dad().get_id()
		proband_mum = proband.get_mum().get_id()
		
		affected = self.family.get_affected_family_members()
		unaffected = self.family.get_unaffected_family_members()
				
		for gene in self.candidate_compound_het_dict:
			
			# If we have more than one candidate in the gene
			if len(self.candidate_compound_het_dict[gene]) > 1:
				
				# Get all candidate combinations within gene
				all_combinations = list(itertools.combinations(self.candidate_compound_het_dict[gene],2))
				
				# Loop Through each pair
				for pair in all_combinations:

					if compound_het_pair_pass_filter(pair,
													 affected,
													 unaffected,
													 proband_dad,
													 proband_mum,
													 include_denovo=include_denovo,
													 allow_hets_in_unaffected=allow_hets_in_unaffected,
													 check_affected=check_affected) == True:
						
						if gene not in self.filtered_compound_het_dict:
							
							self.filtered_compound_het_dict[gene] = [pair]
							
						else:


							self.filtered_compound_het_dict[gene].append(pair)
						

	def get_filtered_compound_hets_as_dict(self):
		"""
		Flatten the filtered_compound_het_dict. Variants in the self.filtered_compound_het_dict will be added to 
		a flat self.final_compound_hets dict.

		Use for trio analysis.

		Input: Self

		Returns: None - Loads self.final_compound_hets

		"""

		self.final_compound_hets = {}

		assert self.filtered_compound_het_dict != False

		for gene in self.filtered_compound_het_dict:

			for pair in self.filtered_compound_het_dict[gene]:

				for variant in pair:

					self.final_compound_hets[variant.variant_id] = None


	def filter_variants(self, function, args):
		"""
		Apply a filtering function to the self.variant_dict.

		Input:

			function: (function) A Python function that should take a Variant object as its first argument.
			args: (Tuple) A tuple of additional arguments to the function argument e.g. (1,2,)

		Returns:

			None - filters self.variant_dict

		"""

		self.variant_dict = {k : v for k,v in filter(lambda x: function(x[1], *args), self.variant_dict.items())}


	def to_df(self, add_inheritance=True,
				 lenient=False,
				 low_penetrance_genes={},
				 min_parental_gq_dn=30,
				 min_parental_depth_dn=10,
				 max_parental_alt_ref_ratio_dn=0.04,
				 min_parental_gq_upi=30,
				 min_parental_depth_upi=10):
		"""
		Convert variant_dict to Pandas DataFrame.

		Input: None

		Returns: Pandas DataFrame object.

		"""

		df_list = []

		for variant in self.variant_dict :
    
		    var = self.variant_dict [variant]
		    
		    for transcript in var.transcript_annotations:
		        
		        row = {}
		        
		        row['chromosome'] = var.chrom
		        row['position'] = var.pos
		        row['ref'] = var.ref
		        row['alt'] = var.alt
		        row['filter_status'] = '|'.join(var.filter_status)
		        row['family_id'] = var.family.family_id
		        row['variant_id'] = var.variant_id

		        if add_inheritance == True:

		        	row['inheritance_models'] = '|'.join(var.get_matching_inheritance_models(compound_het_dict= self.final_compound_hets,
		        																			lenient=lenient,
		        																			min_parental_gq_dn=min_parental_gq_dn,
		        																			min_parental_depth_dn =min_parental_depth_dn,
		        																			max_parental_alt_ref_ratio_dn= max_parental_alt_ref_ratio_dn,
		        																			min_parental_gq_upi = min_parental_gq_upi,
		        																			min_parental_depth_upi = min_parental_depth_upi
		        																			))

		        row['worst_consequence'] = var.get_worst_consequence()
		        
		        for sample in var.genotypes:
		                   
		            row[f'{sample}_GT'] = '/'.join(var.genotypes[sample]['genotype'])
		            row[f'{sample}_AD'] =  ','.join(str(x) for x in var.genotypes[sample]['allele_depths'])
		            row[f'{sample}_DP'] = var.genotypes[sample]['depth']
		            row[f'{sample}_GQ'] = var.genotypes[sample]['genotype_quality']
		            
		        for info_annotation in var.info_annotations:
		            
		            row[f'info_{info_annotation}'] = var.info_annotations[info_annotation]
		            
		        for transcript_annotation in transcript:
		            
		            row[f'csq_{transcript_annotation}'] = transcript[transcript_annotation]
		        
		        
		        df_list.append(row)

		df = pd.DataFrame(df_list)

		return df

















import os
import csv
import glob

def run_phenolyzer(perl_executable, phenolyzer_script, temp_dir, job_name, hpo_terms, symbol=True):
	"""
	Python wrapper for running Phenolyzer on default settings.

	Input:

		perl_executable (String): Location of the perl executable.
		phenolyzer_script (String): The location of the disease_annotation.pl script.
		temp_dir (String): A temporary directory to store results.
		job_name (String): A name for the temporary files.
		hpo_terms (List): List of HPO terms to use.
		symbol (Boolean): Whether to use the gene symbol or the gene_id as the key in the output dictionary.

	Output:
		
		gene_score_dict (Dict): A dictionary with the gene as the key and the score as the value.

	"""

	assert temp_dir != '' and job_name != ''

	final_gene_list = f'{temp_dir}/phenolyser_result_{job_name}.final_gene_list'

	gene_score_dict = {}

	for term in hpo_terms:

		if term[0:3] != 'HP:':

			raise ValueError(f'{term} does not begin with HP:')


	hpo_terms = ';'.join(hpo_terms)
	command = f'{perl_executable} {phenolyzer_script} "{hpo_terms}" -p -ph -logistic -out {temp_dir}/phenolyser_result_{job_name}'

	os.system(command)

	exists = os.path.isfile(final_gene_list)

	if not exists:

		raise ValueError('No output file created.')

	with open(final_gene_list) as csvfile:

		reader = csv.reader(csvfile, delimiter='\t')

		for row in reader:

			if row[1] != 'Gene':

				if symbol:

					gene_score_dict[row[1]] = row[3]

				else:

					gene_score_dict[row[2]] = row[3]


	# Remove old files.

	file_roots = f'{temp_dir}/phenolyser_result_{job_name}'

	files = glob.glob(f'{file_roots}*')

	for file in files:

		os.remove(file)

	return gene_score_dict

def run_phen2gene(phen2gene_dir, temp_dir, job_name, hpo_terms, symbol=True):
	"""
	Python wrapper for running Phen2Gene on default settings.

	Input:

		phen2gene_dir (String): Location of the directory containing the Phen2Gene script.
		temp_dir (String): A temporary directory to store results.
		job_name (String): A name for the temporary files.
		hpo_terms (List): List of HPO terms to use.
		symbol (Boolean): Whether to use the gene symbol or the gene_id as the key in the output dictionary.

	Output:
		
		gene_score_dict (Dict): A dictionary with the gene as the key and the score as the value.

	"""

	assert temp_dir != '' and job_name != ''

	gene_score_dict = {}

	hpo_terms = ' '.join(hpo_terms)

	command = f'cd {phen2gene_dir} && python {phen2gene_dir}/phen2gene.py -m  {hpo_terms} -out temp/{job_name}'
	os.system(command)

	final_gene_list = f'{phen2gene_dir}/temp/{job_name}/input_case.final_candidate_gene_list'

	exists = os.path.isfile(final_gene_list)

	if not exists:

		raise ValueError('No output file created.')

	with open(final_gene_list) as csvfile:

		reader = csv.reader(csvfile, delimiter='\t')

		for row in reader:

			if row[1] != 'Gene':

				if symbol:

					gene_score_dict[row[1]] = row[3]

				else:

					gene_score_dict[row[2]] = row[3]


	os.remove(final_gene_list)

	return gene_score_dict














def parse_csq_field(csq_fields, field_description):
	"""
	Get the VEP CSQ field as a list of dictionaries.

	Each transcript will be its own dictionary in the list.

	Input:

	csq_fields - A tuple containing the different transcipt annotations for VEP e.g ('A|missense_variant', 'A|5_prime_UTR_variant' )

	field_description - List of CSQ field descriptions from VCF header e.g. ['Allele', 'Consequence']

	Returns:

		A list of dictionaries - each transcript the variant falls within has its own dictionary. The key to each dictionary comes from the field_description input.

	"""

	consequence_list = []

	for csq in csq_fields:

		csq_dict = {}

		csq = csq.split('|')

		for key, value in zip(field_description, csq):

			csq_dict[key] = value

		consequence_list.append(csq_dict)

	return consequence_list


def get_info_field_dict(info_fields, vep_csq_key='CSQ'):
	"""
	Get the info fields from a VCF as a dictionary.

	Exclude the VEP annotation field e.g. CSQ.

	Input:

	info_fields - The info_fields object from the PySam VCF parser.
	vep_csq_key - The key to access the CSQ field.


	"""

	info_dict = {}

	for key in info_fields.keys():

		if key != vep_csq_key:

			if isinstance(info_fields[key], tuple) and len(info_fields[key]) ==1:

				info_dict[key] = info_fields[key][0]

			else:

				info_dict[key] = info_fields[key]


	return info_dict

def compound_het_pair_pass_filter(pair,
							 affected,
							 unaffected,
							 proband_dad,
							 proband_mum,
							 include_denovo=True,
							 allow_hets_in_unaffected=False,
							 check_affected=True):

	"""
	Filter compound hets in a proband with both parents. Return True if they are genuine compoun hets.

	Input:

	pair - a list containing a compound het pair
	affected - a list containing affected family members.
	unaffected - A list containing unaffected family members.

	Output:

	True - If the pair pass the compound het filter
	False - If the pair fail the compound het filter.

	Rules:

	1) No unaffected samples can have the pair of variants. Can be adjusted using the allow_hets_in_unaffected argument.
	2) All affected samples must have the pair of variants or be missing the genotype data. Can be adjusted using the check_affected argument.
	3) a) One of the pair must be inherited from mum and the other from dad.
	   b) If include_denovo is True then one of the pair can be de_novo and the other inherited from either parent or both can be de_novo. There are no minimum requirements e.g. depth on the de_novo calls.

	"""

	var1 = pair[0]
	var2 = pair[1]

	# Check unaffected  - if unaffected people have pair then break
	if allow_hets_in_unaffected == False:

		for sample_id in unaffected:

			if var1.is_het(sample_id) and var2.is_het(sample_id):

				return False

	# Check affected - all affected must have the variant pair or be missing the variant
	if check_affected == True:

		for sample_id in affected:

			if var1.is_hom_ref(sample_id) or var2.is_hom_ref(sample_id) and (var1.is_missing(sample_id) == False or var2.is_missing(sample_id) == False) :

				return False

	# if var 1 comes from mum and var 2 comes from dad
	if var1.has_no_alt(proband_dad) and var1.is_het(proband_mum) and var2.has_no_alt(proband_mum) and var2.is_het(proband_dad): 
		return True

	# if var 2 comes from mum and var 1 comes from dad
	elif var1.has_no_alt(proband_mum) and var1.is_het(proband_dad) and var2.has_no_alt(proband_dad) and var2.is_het(proband_mum):
		return True

	# mum is het for var1 and hom ref for var2 and dad is missing both genotypes for var 1 and var 2
	elif var1.is_het(proband_mum) and var2.is_hom_ref(proband_mum) and var1.is_missing(proband_dad) and var2.is_missing(proband_dad):
		return True

	# dad is het for var1 and hom ref for var2 and mum is missing both genotypes for var 1 and var 2
	elif var1.is_het(proband_dad) and var2.is_hom_ref(proband_dad) and var1.is_missing(proband_mum) and var2.is_missing(proband_mum):
		return True

	# if var1 is denovo and var2 is het in dad
	elif (var1.matches_denovo(min_parental_gq=0, min_parental_depth=-1, max_parental_alt_ref_ratio=1) and
		  var2.is_het(proband_dad) and
		  var2.has_no_alt(proband_mum) and
		  include_denovo == True):
		return True

	# if var1 is denovo and var2 is het in mum
	elif (var1.matches_denovo(min_parental_gq=0, min_parental_depth=-1, max_parental_alt_ref_ratio=1) and
		  var2.is_het(proband_mum) and
		  var2.has_no_alt(proband_dad) and
		  include_denovo == True):
		return True

   # if var2 is denovo  and var1 is het in mum
	elif (var2.matches_denovo(min_parental_gq=0, min_parental_depth=-1, max_parental_alt_ref_ratio=1) and
		  var1.is_het(proband_mum) and
		  var1.has_no_alt(proband_dad) and
		  include_denovo == True):
		return True

   # if var2 is denovo  and var1 is het in dad
	elif (var2.matches_denovo(min_parental_gq=0, min_parental_depth=-1, max_parental_alt_ref_ratio=1) and
		  var1.is_het(proband_dad) and
		  var1.has_no_alt(proband_mum) and
		  include_denovo == True):
		return True

	# if both var1 and var2 are denovo
	elif (var1.matches_denovo(min_parental_gq=0, min_parental_depth=-1, max_parental_alt_ref_ratio=1) and
		  var2.matches_denovo(min_parental_gq=0, min_parental_depth=-1, max_parental_alt_ref_ratio=1) and
		  include_denovo == True):
		return True

	# Options below will only be True if allow_hets_in_unaffected is True

	# if mum is het for var1 and dad is het for var1 and var2
	elif var1.is_het(proband_mum) and var2.has_no_alt(proband_mum) and var1.is_het(proband_dad) and var2.is_het(proband_dad):
		return True
	
	# if mum is het for var2 and dad is het for var1 and var2
	elif var2.is_het(proband_mum) and var1.has_no_alt(proband_mum) and var1.is_het(proband_dad) and var2.is_het(proband_dad):
		return True

	# if dad is het for var1 and mum is het for var1 and var2
	elif var1.is_het(proband_dad) and var2.has_no_alt(proband_dad) and var1.is_het(proband_mum) and var2.is_het(proband_mum):
		return True
	
	# if dad is het for var2 and mum is het for var1 and var2
	elif var2.is_het(proband_dad) and var1.has_no_alt(proband_dad) and var1.is_het(proband_mum) and var2.is_het(proband_mum):
		return True

	# if both parents are het for both var1 and var2
	elif var1.is_het(proband_dad) and var2.is_het(proband_dad) and var1.is_het(proband_mum) and var2.is_het(proband_mum):
		return True 

	# if mum is het for both var1 and var2 and dad is missing both var1 and var2
	elif var1.is_het(proband_mum) and var2.is_het(proband_mum) and var1.is_missing(proband_dad) and var2.is_het(proband_dad):
		return True

	# if dad is het for both var1 and var2 and mum is missing both var1 and var2
	elif var1.is_het(proband_dad) and var2.is_het(proband_dad) and var1.is_missing(proband_mum) and var2.is_het(proband_mum):
		return True
	else:
		return False



		


from pyvariantfilter.family_member import FamilyMember
import csv

class Family:
	"""
	A Family class for holding information



	"""
	
	def __init__(self, family_id):
		
		self.family_id = family_id
		self.family_members = []
		
	def __repr__(self):
		 return self.family_id
		
	def add_family_member(self, family_member):
		"""
		Add a FamilyMember to an Family.

		Input:

			family_member(FamilyMember :  Parents must already be family_members.

		Returns:

			None

		"""
		assert isinstance(family_member, FamilyMember)

		all_family_members = self.get_all_family_member_ids()

		if family_member.dad != None:

			if family_member.dad.get_id() not in all_family_members:

				raise ValueError(f'{family_member.get_id()} not in Family.')

		if family_member.mum != None:

			if family_member.mum.get_id()  not in all_family_members:

				raise ValueError(f'{family_member.get_id()} not in Family.')

		for existing_member in self.family_members:
			
			if family_member.family_member_id == existing_member.family_member_id:
				
				raise ValueError(f'Family member ({family_member.family_member_id}) already in Family.')
		
		self.family_members.append(family_member)
		
	def get_affected_family_members(self):
		"""
		Return those family members who are affected.

		Input: Self

		Returns:

			list: List of affected family member ids.
		"""
		
		return [family_member.get_id() for family_member in self.family_members if family_member.affected == True]
	
	def get_unaffected_family_members(self):
		"""
		Return those family members who are unaffected.

		Input: Self

		Returns:

			list: List of unaffected family member ids.
		"""
		
		return [family_member.get_id() for family_member in self.family_members if family_member.affected == False]
	
	def get_male_family_members(self):
		"""
		Return Male (1) family members.

		Input: Self

		Returns:

			list: List of Male (1) family member ids.

		"""
		return [family_member.get_id() for family_member in self.family_members if family_member.sex == 1]
	
	def get_female_family_members(self):
		"""
		Return Female (2) family members.

		Input: Self

		Returns:

			list: List of Male (2) family member ids.

		"""
		return [family_member.get_id() for family_member in self.family_members if family_member.sex == 2]
	
	def get_all_family_members(self):
		"""
		Return all family members objects in the family.

		Input: Self

		Returns:

			list: List of FamilyMember objects in the family.

		"""
		
		return self.family_members

	def get_all_family_member_ids(self):
		"""
		Returns the ids of all FamilyMember objects in the family.

		Input: Self

		Returns:

			list: List of all FamilyMember ids in the family.

		"""
		
		return [family_member.get_id() for family_member in self.get_all_family_members()]
	
	def set_proband(self, family_member_id):
		"""
		Set a specific sample as the proband.

		Input:

			family_member_id (String) : The family_member_id attribute of a FamilyMember within the family.

		Returns:

			None

		"""

		
		for existing_member in self.family_members:
			
			# Remove existing proband attribute
			existing_member.proband = False
			
			if existing_member.get_id() == family_member_id:

				assert existing_member.affected == True
				
				existing_member.proband = True
				
				return
		
		# If the user tries to assign a FamilyMember not in the family.
		raise ValueError(f'No Family Member with id: {family_member_id}')
		
	def get_proband(self):
		"""
		Return the proband within the Family. Only one sample can be the proband.

		Input: Self

		Returns:

			proband (FamilyMember): The FamilyMember object instance of the proband in the Family.

		"""
		
		probands = [family_member for family_member in self.family_members if family_member.proband == True]

		assert len(probands) == 1

		return probands[0]


	def proband_has_both_parents(self):
		"""
		Return whether the proband has both parents availible?

		Input: Self

		Returns:

			True - If proband has a mum and dad
			False - Otherwise.

		"""

		proband = self.get_proband()

		if proband.mum != None and proband.dad != None:

			return True

		return False


	def get_proband_id(self):
		"""
		Return the family_member_id within the Family. Only one sample can be the proband.

		Input: Self

		Returns:

			family_member_id (String): The FamilyMember object instance of the proband in the Family.

		"""
		
		probands = [family_member for family_member in self.family_members if family_member.proband == True]

		assert len(probands) == 1

		return probands[0].get_id()
	
	def get_affected_female_members(self):
		"""
		Returns all affected Female (2) members of the Family.

		Input: Self

		Returns:

			list: A list containing the family_member_ids of the affected Female members of the Family.

		"""
		return [family_member.get_id() for family_member in self.family_members if family_member.sex == 2 and family_member.affected == True]
	
	def get_unaffected_female_members(self):
		"""
		Returns all unaffected Female (2) members of the Family.

		Input: Self

		Returns:

			list: A list containing the family_member_ids of the unaffected Female members of the Family.

		"""
		return [family_member.get_id() for family_member in self.family_members if family_member.sex == 2 and family_member.affected == False]
	
	def get_affected_male_members(self):
		"""
		Returns all affected Male (1) members of the Family.

		Input: Self

		Returns:

			list: A list containing the family_member_ids of the affected Memale members of the Family.

		"""
		
		return [family_member.get_id() for family_member in self.family_members if family_member.sex == 1 and family_member.affected == True]
	
	def get_unaffected_male_members(self):
		"""
		Returns all unaffected Male (1) members of the Family.

		Input: Self

		Returns:

			list: A list containing the family_member_ids of the unaffected Memale members of the Family.

		"""
		
		return [family_member.get_id() for family_member in self.family_members if family_member.sex == 1 and family_member.affected == False]
	
	def get_daughters(self, family_member_id):
		"""
		Get the daughters of a FamilyMember object.

		Input:

			family_member_id (String): family_member_id of a FamilyMember object in the family.

		Returns:

			daughters (List) : List of daughter FamilyMember's  belonging to the FamilyMember object specified by family_member_id.

		"""
		
		daughters = []

		assert family_member_id in self.get_all_family_member_ids()
		
		for family_member in self.family_members: 
						
			if family_member.dad != None:
						
				if family_member.dad.get_id() == family_member_id and family_member.sex == 2:
				
					daughters.append(family_member)
					
			if family_member.mum != None:
								
				if family_member.mum.get_id() == family_member_id and family_member.sex == 2:
				
					daughters.append(family_member)           
				
		return daughters

	def get_daughter_ids(self, family_member_id):
		"""
		Get the daughters of a FamilyMember object.

		Input:

			family_member_id (String): family_member_id of a FamilyMember object in the family.

		Returns:

			daughters (list): List of daughter family_member_ids  belonging to the FamilyMember object specified by family_member_id.

		"""
		
		daughters = []

		assert family_member_id in self.get_all_family_member_ids()
		
		for family_member in self.family_members: 
						
			if family_member.dad != None:
						
				if family_member.dad.get_id() == family_member_id and family_member.sex == 2:
				
					daughters.append(family_member.get_id())
					
			if family_member.mum != None:
								
				if family_member.mum.get_id() == family_member_id and family_member.sex == 2:
				
					daughters.append(family_member.get_id())           
				
		return daughters

	
	def get_sons(self, family_member_id):
		"""
		Get the sons of a FamilyMember object.

		Input:

			family_member_id (String): family_member_id of a FamilyMember object in the family.

		Returns:

			sons (List): List of son family_member_ids belonging to the FamilyMember object specified by family_member_id.

		"""
		
		sons = []

		assert family_member_id in self.get_all_family_member_ids()
		
		for family_member in self.family_members: 
						
			if family_member.dad != None:
						
				if family_member.dad.get_id() == family_member_id and family_member.sex ==1:
				
					sons.append(family_member)
					
			if family_member.mum != None:
				
				if family_member.mum.get_id() == family_member_id and family_member.sex ==1:
				
					sons.append(family_member)           
				
		return sons

	def get_son_ids(self, family_member_id):
		"""
		Get the sons of a FamilyMember object.

		Input:

			family_member_id (String): family_member_id of a FamilyMember object in the family.

		Returns:

			sons (List) : List of son family_member_ids belonging to the FamilyMember object specified by family_member_id.

		"""
		
		sons = []

		assert family_member_id in self.get_all_family_member_ids()
		
		for family_member in self.family_members: 
						
			if family_member.dad != None:
						
				if family_member.dad.get_id() == family_member_id and family_member.sex ==1:
				
					sons.append(family_member.get_id())
					
			if family_member.mum != None:
				
				if family_member.mum.get_id() == family_member_id and family_member.sex ==1:
				
					sons.append(family_member.get_id())           
				
		return sons

	def read_from_ped_file(self, ped_file_path, family_id, proband_id):
		"""
		Populate a family object with information from a PED file.

		See https://gatkforums.broadinstitute.org/gatk/discussion/7696/pedigree-ped-files

		Input:

			ped_file_path (String): Path to the PED file.
			family_id (String): Family to extract from the PED file.
			proband_id (String): The ID of the sample in the PED file to set as the proband.

		Returns:

			None - Populates the Family object with PED information.

		"""

		family_dict = {}

		with open(ped_file_path, 'r') as csvfile:

			pedreader = csv.reader(csvfile, delimiter='\t')

			for row in pedreader:
 
				if row[0] == family_id:

					if row[5] == '2':

						affected = True

					elif row[5] == '1':

						affected = False

					new_family_member = FamilyMember(family_member_id=row[1],family_id=family_id, sex=int(row[4]), affected=affected)

					family_dict[new_family_member.get_id()] = new_family_member

					self.add_family_member(new_family_member)


		with open(ped_file_path, 'r') as csvfile:

			pedreader = csv.reader(csvfile, delimiter='\t')

			for row in pedreader:

				if row[2] in family_dict:

					family_dict[row[1]].dad = family_dict[row[2]]

				if row[3] in family_dict:

					family_dict[row[1]].mum = family_dict[row[3]]


		self.set_proband(proband_id)







from pyvariantfilter.family import Family
import statistics

class Variant:
	
	def __init__(self, chrom, pos, ref, alt, filter_status=None, quality=None):
		
		self.chrom = chrom
		self.pos = pos
		self.ref = ref
		self.alt = alt
		self.filter_status = filter_status
		self.quality = quality
		self.info_annotations = None
		self.transcript_annotations = None
		self.variant_id = f'{self.chrom}:{self.pos}{self.ref}>{self.alt}'
		self.family = None
		self.genotypes = {}

		self.is_valid()
			
	def __repr__(self):
		 return self.variant_id

	def is_valid(self):
		"""
		Check the newly constructed Variant is valid.

		Returns:

			None

		"""

		valid_chroms = {'1': None, '2': None, '3': None, '4': None,
		 '5': None, '6': None, '7': None, '8': None, '9': None,
		 '10': None, '11': None, '12': None, '13': None,
		 '14': None, '15': None, '16': None, '17': None,
		 '18': None, '19': None, '20': None, '21': None,
		 '22': None, 'X': None, 'Y': None, 'MT': None, 'M': None}

		# Chromosome is correct
		if self.chrom not in valid_chroms:

			raise ValueError(f'Chromosome not in {valid_chroms} for variant {self.variant_id}.')

		# Position must be an Integer
		if isinstance(self.pos, int) == False:

			raise ValueError('Position not an integer.')

		# ref and alt must be an Integer
		if isinstance(self.ref, str) == False or isinstance(self.alt, str) == False:

			raise ValueError('The Ref and Alt must be strings.')

		 
	def add_genotype(self, family_member_id, genotype, allele_depths, genome_quality, depth):
		"""
		Add a genotype to a Varaint object.

		Input:

			family_member_id: family_member_id from the FamilyMember object. (String)
			genotype: List of genotypes e.g ['A', 'T'] (List of Strings)
			allele_depths List of allele_depths - First item in list should be ref and second alt e.g. [10, 5] (List of Integers)
			genome_quality - The genome quality (GQ) score. Integer

		Returns:

			None
		"""

		assert isinstance(self.family, Family) == True
		assert family_member_id in self.family.get_all_family_member_ids()
		assert isinstance(genotype, list)
		assert len(genotype) == 2
		assert genotype[0] in [self.ref, self.alt, '.', '*']
		assert genotype[1] in [self.ref, self.alt, '.', '*']
		assert isinstance(allele_depths, list)
		assert len(allele_depths) == 2
		assert isinstance(allele_depths[0], int)
		assert isinstance(allele_depths[1], int)
		assert isinstance(depth, int)
		assert isinstance(genome_quality, int)

		self.genotypes[family_member_id] = {'genotype': genotype,
									'allele_depths': allele_depths,
									'genotype_quality': genome_quality,
									'depth': depth}
		  
	def add_family(self, family):
		"""
		Associate a family object with this variant.

		Input:

			family: Family object instance

		Returns:

			None
		
		"""
		assert isinstance(family, Family)
	
		self.family = family
		
	def get_genes(self, feature_key='Feature'):
		"""
		Get the genes associated with a variant.

		This function can be used to get any features associated with a variant by changing the feature_key.

		Input: feature_key - A key for each transcript in the self.transcript_annotations list

		Returns: genes - A list of all the features/genes in all transcripts.

		"""

		genes = []

		if self.transcript_annotations != None:

			for csq_dict in self.transcript_annotations:

				genes.append(csq_dict[feature_key])

		else:

			raise ValueError('No gene annotations present.')

		return genes
	
	def get_worst_consequence(self, consequence_severity=[
		'transcript_ablation', 
		'splice_acceptor_variant', 
		'splice_donor_variant', 
		'stop_gained', 
		'frameshift_variant',
		'stop_lost',
		'start_lost',
		'transcript_amplification',
		'inframe_insertion',
		'inframe_deletion',
		'missense_variant',
		'protein_altering_variant',
		'splice_region_variant',
		'incomplete_terminal_codon_variant',
		'start_retained_variant',
		'stop_retained_variant',
		'synonymous_variant',
		'coding_sequence_variant',
		'mature_miRNA_variant',
		'5_prime_UTR_variant',
		'3_prime_UTR_variant',
		'non_coding_transcript_exon_variant',
		'intron_variant',
		'NMD_transcript_variant',
		'non_coding_transcript_variant',
		'upstream_gene_variant',
		'downstream_gene_variant',
		'TFBS_ablation',
		'TFBS_amplification',
		'TF_binding_site_variant',
		'regulatory_region_ablation',
		'regulatory_region_amplification',
		'feature_elongation',
		'regulatory_region_variant',
		'feature_truncation',
		'intergenic_variant'], consequence_key='Consequence'):
		"""
		Get the worst consequence of all the transcripts in the self.transcript_annotations.

		The order of severity can be changed using the consequence_severity variable.

		Input:

		consequence_severity: list of consequences from worst to best.
		consequence_key: The key to use to look for the consequence. 
		


		"""
		
		if self.transcript_annotations == None:

			raise ValueError('No transcript annotations present.')

		else:

			consequences = []

			for csq_dict in self.transcript_annotations:

				if consequence_key not in csq_dict:

					raise ValueError(f'The consequence key ({consequence_key}) does not exist in the transcript annotations.')

				consequences.append(csq_dict[consequence_key])

			worst_index = 9999
			
			for consequence in consequences:
				
				if consequence == None:
					
					break
				
				split_consequences = consequence.split('&')
				
				for c in split_consequences:
				
					index = consequence_severity.index(c)
				
					if index < worst_index:
					
						worst_index = index
			
			if worst_index == 9999:
				
				return None
			
			else:
					
				return consequence_severity[worst_index]


		
	def is_hom_ref(self, family_member_id):
		"""
		Is a variant genotype homozygous reference?

		Input:

			family_member_id: family_member_id of a Family Member object.

		Returns:

			True: If family_member_id is homozygous reference for the variant.
			False: If family_member_id is not homozygous reference for the variant.

		"""
		
		gt = self.genotypes[family_member_id]['genotype']
		
		if gt.count(self.ref) == 2:
			
			return True
		
		else:
			
			return False
		
	def has_no_alt(self, family_member_id):
		"""
		Does the alt allele not exist in the genotype?

		Input:

			family_member_id: family_member_id of a Family Member object.

		Returns:

			True: If family_member_id does not have the alt allele.
			False: If family_member_id does have the alt allele.

		"""
		
		gt = self.genotypes[family_member_id]['genotype']
			
		if gt.count(self.alt) == 0:
			
			return True
		
		else:
			
			return False
		
	
	def is_het(self, family_member_id):
		"""
		Does the alt allele have a count of one in the genotype?

		Input:

			family_member_id: family_member_id of a Family Member object.

		Returns:

			True: If family_member_id has an alt count of one.
			False: If family_member_id does not have an alt count of one.

		"""
		
		gt = self.genotypes[family_member_id]['genotype']
		
		if gt.count(self.alt) == 1:
			
			return True
		
		else:
			
			return False
		
	def has_alt(self, family_member_id):
		"""
		Does the alt allele exist in the genotype?

		Input:

			family_member_id: family_member_id of a Family Member object.

		Returns:

			True: If family_member_id genotype has the alt.
			False: If family_member_id does not have the alt.

		"""
		
		gt = self.genotypes[family_member_id]['genotype']
		
		if gt.count(self.alt) == 0:
			
			return False
		
		else:
			
			return True
	
	def is_hom_alt(self, family_member_id):
		"""
		Does the alt allele have a count of two?

		Input:

			family_member_id: family_member_id of a Family Member object.

		Returns:

			True: If family_member_id genotype has an alt count of two.
			False: If family_member_id does not have an alt count of two.

		"""
		
		gt = self.genotypes[family_member_id]['genotype']
		
		if gt.count(self.alt) == 2:
			
			return True
		
		else:
			
			return False

	def is_homozygous(self, family_member_id):
		"""
		Is the family member homozygous for either the reference or alt allele?

		Input:

			family_member_id: family_member_id of a Family Member object.

		Returns:

			True: If family_member_id genotype has an allele count of 2 for either the reference or alt allele.
			False: If family_member_id genotype does not have an allele count of 2 for either the reference or alt allele.

		"""

		if self.is_hom_ref(family_member_id) == True:

			return True

		if self.is_hom_alt(family_member_id) == True:

			return True

		return False

	
	def is_missing(self, family_member_id):
		"""
		Are both alleles missing i.e ./. in VCF form.

		Input:

			family_member_id: family_member_id of a Family Member object.

		Returns:

			True: if the genotype is missing.
			False: if the genotype is not missing.

		"""
		
		gt = self.genotypes[family_member_id]['genotype']
		
		if gt.count('.') == 2:
			
			return True
		
		else:
			
			return False
		
	def is_mixed(self, family_member_id):
		"""
		Is one allele missing i.e 1/. in VCF form?

		Input:

			family_member_id: family_member_id of a Family Member object.

		Returns:

			True: if the genotype is mixed.
			False: if the genotype is not mixed.

		"""
		
		gt = self.genotypes[family_member_id]['genotype']
		
		if gt.count('.') == 1:
			
			return True
		
		else:
			
			return False
	
	def get_depth(self, family_member_id):
		"""
		Get the genotype depth for a family_member.

		Input:

			family_member_id: family_member_id of a Family Member object.

		Returns:

			ad: genotype depth

		"""
		
		ad = self.genotypes[family_member_id]['depth']
		
		return ad
	
	def get_genotype_quality(self, family_member_id):
		"""
		Get the genotype quality for a family_member.

		Input:

			family_member_id: family_member_id of a Family Member object.

		Returns:

			gq: genotype quality

		"""

		
		gq = self.genotypes[family_member_id]['genotype_quality']
		
		return gq
	
	def get_alt_reads(self, family_member_id):
		"""
		Get the alt read count for a family_member.

		Input:

			family_member_id: family_member_id of a Family Member object.

		Returns:

			ad: alt read count

		"""	
		ad = self.genotypes[family_member_id]['allele_depths']
		
		return ad[1]

	def get_ref_reads(self, family_member_id):
		"""
		Get the ref read count for a family_member.

		Input:

			family_member_id: family_member_id of a Family Member object.

		Returns:

			ad: ref read count

		"""	
		ad = self.genotypes[family_member_id]['allele_depths']
		
		return ad[0]
	
	def matches_autosomal_dominant(self, lenient=False, low_penetrance_genes={}):
		"""
		Does the variant match an autosomal dominant inheritance pattern?
				
		Input:

			lenient: (Boolean) - allow affected who are not the proband to be homozygous alternate
			low_penetrance_genes: (Dict) - Dictionary of low penetrance genes - \
			variants within genes in this dictionary are allowed to have unaffected  \
			people with the variant.

		Returns:

			True - If matches autosomal dominant inheritance pattern.
			False - If does not match autosomal dominant inheritance pattern.


		Rules:

			1) Variant must be on an autosome.
			2) All affected samples must be heterozygous or missing e.g. ./. See lenient option to allow homozygous alternate genotypes in affected samples other than the proband.
			3) If the variant is not in a low penetrant gene then all unaffected samples must be homozygous reference or have a missing genotype.

		"""

		if self.chrom in ['X', 'Y', 'MT', 'M']:
			
			return False
	
		
		# Check that all affected are het
		affected = self.family.get_affected_family_members()
		
		if lenient == False:
		
			for family_member_id in affected:

				if self.is_het(family_member_id) == False and self.is_missing(family_member_id) == False:

					return False
				
		else:
			
			# affected people who are not the proband are allowed to be hom alt
			for family_member_id in affected:
					
				if self.family.get_proband_id() != family_member_id:
					
					if self.has_alt(family_member_id) == False and self.is_missing(family_member_id) == False:

						return False
					
				else:
										
					if self.is_het(family_member_id) == False and self.is_missing(family_member_id) == False:

						return False

		check_unaffected = True

		if low_penetrance_genes:

			genes = self.get_genes(feature_key='SYMBOL')

			for gene in genes:

				if gene in low_penetrance_genes:

					check_unaffected = False
					break

		if check_unaffected == True:

			# Check that all unaffected are hom ref and not missing if they are return false
			unaffected = self.family.get_unaffected_family_members()
				
			for family_member_id in unaffected:
					
				if self.has_alt(family_member_id) == True:
						
					return False
				
		return True
		
	
	def matches_autosomal_reccessive(self):
		"""
		Does the variant match an autosomal reccessive inheritance pattern?   
		
		Input: Self

		Returns:

			True - If matches autosomal reccessive inheritance pattern.
			False - If does not match autosomal reccessive inheritance pattern.

		Rules:

		1) Variant must be on an Autosome.
		2) All affected samples must be homozygous for the alternate allele. Can be missing.
		3) No unaffected samples can be homozygous for the alternate allele. Can be missing.

		"""
		
		if self.chrom in ['X', 'Y', 'MT', 'M']:
			
			return False
		
		# Check that all affected are hom alt
		affected = self.family.get_affected_family_members()
		
		for family_member_id in affected:
					
			if self.is_hom_alt(family_member_id) == False and self.is_missing(family_member_id) == False:
									
				return False
		
		# Check that unaffected are not hom alt 
		unaffected = self.family.get_unaffected_family_members()
			
		for family_member_id in unaffected:
							
			if self.is_hom_alt(family_member_id) == True:
					
				return False      
			
		return True
		
	
	def matches_denovo(self, min_parental_gq=30, min_parental_depth=10, max_parental_alt_ref_ratio=0.04):
		"""
		Given a sample as a proband did the variant arise de novo?

		Input:

			min_parental_gq: int - Both parental genotypes must have a GQ value above this.
			min_parental_depth: int - Both parental genotypes must have a depth above this.
			max_parental_alt_ref_ratio: int - Parents cannot have an alt/ref ratio higher than this.

		Returns:

			True - If matches de novo inheritance pattern.
			False - If does not match de novo autosomal inheritance pattern. 

		Rules:

			1) Variant in proband and not in either parent.
			2) Parents must have a GQ value above min_parental_gq
			3) Parents must have a DP value above min_parental_depth
			4) Parents must have a alt/ref ratio below max_parental_alt_ref_ratio

		
		"""
		proband = self.family.get_proband()
		
		mum = proband.get_mum()
		dad = proband.get_dad()

		if self.has_alt(proband.get_id()):
			
			# Must have both parents to calculate
			if mum == None or dad == None:
				
				return False

			else:
			
				mum_id = mum.get_id()
				dad_id = dad.get_id()
				
				mum_gq = self.get_genotype_quality(mum_id)
				mum_dp = self.get_depth(mum_id)
				mum_alt = self.get_alt_reads(mum_id)
				mum_ref = self.get_ref_reads(mum_id)

				# no div by 0 - what if ads are 0,0
				if mum_ref == 0:

					mum_ad_ratio = 0

				else:

					mum_ad_ratio = mum_alt / mum_ref
				
				dad_gq = self.get_genotype_quality(dad_id)
				dad_dp = self.get_depth(dad_id)
				dad_alt = self.get_alt_reads(dad_id) 
				dad_ref = self.get_ref_reads(dad_id)

				# no div by 0
				if dad_ref == 0:

					dad_ad_ratio = 0

				else:

					dad_ad_ratio = dad_alt / dad_ref
				
				if (self.has_alt(mum_id) == False and 
					mum_gq >= min_parental_gq and 
					mum_dp >= min_parental_depth and 
					mum_ad_ratio <= max_parental_alt_ref_ratio and
					self.has_alt(dad_id) == False and
					dad_gq >= min_parental_gq and
					dad_dp >= min_parental_depth and
					dad_ad_ratio < max_parental_alt_ref_ratio):
					
					return True
				
				else:
					
					return False

		else:

			return False
		
	def matches_x_reccessive(self):
		"""
		Does the variant match an X reccessive inheritance pattern?
		
		Input: Self.

		Returns:

			True - If matches X reccessive inheritance pattern.
			False - If does not match X reccessive inheritance pattern. 

		Rules:

			1) Variant must be on the X chromosome.
			2) All affected female samples must be hom_alt or missing.
			3) No unaffected female samples can be hom_alt.
			4) All affected male samples must have the variant or be missing.
			5) No unaffected male samples can have the variant.

		"""
		
		if self.chrom != 'X':
			
			return False
		
		affected_females = self.family.get_affected_female_members()
		
		for family_member_id in affected_females:
			
			if self.is_hom_alt(family_member_id) == False and self.is_missing(family_member_id) == False:
				
				return False
			
			
		unaffected_females = self.family.get_unaffected_female_members()
		
		for family_member_id in unaffected_females:
			
			if self.is_hom_alt(family_member_id) == True:
				
				return False
	
	
		affected_males = self.family.get_affected_male_members()
		
		for family_member_id in affected_males:
			
			if self.is_hom_ref(family_member_id) == True and self.is_missing(family_member_id) == False:
				
				return False    
	
		unaffected_males = self.family.get_unaffected_male_members()
		
		for family_member_id in unaffected_males:
			
			if self.has_alt(family_member_id) == True:
				
				return False 
			
		return True
	

	def matches_x_dominant(self):
		"""
		Does the variant match an X dominant inheritance pattern?
		
		Input: Self.

		Returns:

			True - If matches X dominant inheritance pattern.
			False - If does not match X dominant inheritance pattern. 

		Rules:

			1) Variant must be on the X chromosome.
			2) The daughters of affected male samples must be affected.
			3) The sons of affected males must not be affected.
			4) Affected male samples must have the variant or be missing.
			5) Affected female samples must be heterozygous or missing.
			6) Unaffected samples must not have the variant.

		"""
		
		# Must be on X chromosome
		if self.chrom != 'X':
			
			return False 
		
		# Check affected males children,
		affected_males = self.family.get_affected_male_members()	
		for family_member_id in affected_males:
			
			# if an affected male has unaffected daughters then no match
			daughters = self.family.get_daughters(family_member_id)
			
			for daughter in daughters:
				
				if daughter.affected == False:
					
					return False
			
			# if an affected male has an affected son then no match
			sons = self.family.get_sons(family_member_id)
			
			for son in sons:
				
				if son.affected == True:
					
					return False
		
		# Affected males must not be hom ref
		for family_member_id in affected_males:
			
			if self.is_hom_ref(family_member_id) == True and self.is_missing(family_member_id) == False :
					
					return False
		
		
		# Affected females must be het or missing.
		affected_females = self.family.get_affected_female_members()
		for family_member_id in affected_females:
			
			if self.is_het(family_member_id) == False and self.is_missing(family_member_id) == False:
				
				return False
			
		# Unaffected samples must not have the variant.
		unaffected = self.family.get_unaffected_family_members()
		for family_member_id in unaffected:
			
			if self.has_alt(family_member_id) == True:
				
				return False 
			
		return True
			
	def matches_uniparental_isodisomy(self, min_parental_gq=30, min_parental_depth=10):
		"""
		Does the variant match a uniparental_isodisomy pattern?
		
		Input:

			min_parental_gq: int - Both parental genotypes must have a GQ value above this.
			min_parental_depth: int - Both parental genotypes must have a depth above this.

		Returns:

			True - If matches uniparental_isodisomy pattern.
			False - If does not match uniparental_isodisomy pattern. 

		"""

		proband = self.family.get_proband()
		mum = proband.get_mum()
		dad = proband.get_dad()

		# Must have both parents to calculate
		if mum == None or dad == None:

			return False

		proband_id = proband.get_id()
		mum_id = mum.get_id()
		dad_id = dad.get_id()
		
		if self.chrom not in ['X', 'Y', 'MT', 'M']:

			if (self.is_hom_alt(proband_id) and (self.is_het(mum_id) and self.has_no_alt(dad_id)) and 
				self.get_depth(mum_id) >= min_parental_depth and self.get_depth(dad_id) >= min_parental_depth and
				self.get_genotype_quality(mum_id) >= min_parental_gq and self.get_genotype_quality(dad_id) >= min_parental_gq):

				return True

			if (self.is_hom_alt(proband_id) and (self.is_het(dad_id) and self.has_no_alt(mum_id)) and 
				self.get_depth(mum_id) >= min_parental_depth and self.get_depth(dad_id) >= min_parental_depth and
				self.get_genotype_quality(mum_id) >= min_parental_gq and self.get_genotype_quality(dad_id) >= min_parental_gq):

				return True


		elif self.is_on_autosome_or_xfemale() == True:

			if (self.is_hom_alt(proband_id) and (self.is_het(mum_id) and self.has_no_alt(dad_id)) and 
				self.get_depth(mum_id) >= min_parental_depth and self.get_depth(dad_id) >= min_parental_depth and
				self.get_genotype_quality(mum_id) >= min_parental_gq and self.get_genotype_quality(dad_id) >= min_parental_gq):

				return True

			if (self.is_hom_alt(proband_id) and (self.has_no_alt(mum_id) and self.has_no_alt(dad_id)) and 
				self.get_depth(mum_id) >= min_parental_depth and self.get_depth(dad_id) >= min_parental_depth and
				self.get_genotype_quality(mum_id) >= min_parental_gq and self.get_genotype_quality(dad_id) >= min_parental_gq):

				return True

		return False

	def matches_compound_het(self, compound_het_dict):
		"""
		Is the variant in the compound het dict?

		Input: compound_het_dict - A dictionary containing variants which are compound hets

		Returns:

			True - If variant is in dict.
			False - If variant not in dict.

		"""

		if self.variant_id in compound_het_dict:

			return True

		else:

			return False

	def matches_mitochrondrial(self):
		"""
		Is the variant on the mitochrondial chromosome?

		Input: self

		Returns:

			True - If variant is on mitochrondial chromosome.
			False - If variant is not on mitochrondial chromosome.

		"""

		if self.chrom =='MT' or self.chrom =='M':

			return True

	def matches_y(self):
		"""
		Is the variant on the Y chromosome?

		Input: self

		Returns:

			True - If variant is on Y chromosome.
			False - If variant is not on Y chromosome.

		"""

		if self.chrom == 'Y':

			return True

	def matches_inheritance_model(self,
								 models,
								 compound_het_dict,
								 lenient=False,
								 low_penetrance_genes={},
								 min_parental_gq_dn=30,
								 min_parental_depth_dn=10,
								 max_parental_alt_ref_ratio_dn=0.03,
								 min_parental_gq_upi=30,
								 min_parental_depth_upi=10):
		"""
		Does the variant match the inputted inheritiance patterns?

		Input:

		models: list of models - possibilities are:

		na - return true ragardless of model.
		autosomal_dominant
		autosomal_reccessive
		x_reccessive
		x_dominant
		de_novo
		uniparental_isodisomy
		mitochrondrial
		y_chrom

		compound_het_dict - Dictionary containing variants which are compound hets
		lenient - In the autosomal_dominant inhertiance pattern whether to allow affected other than the proband to be hom_alt as well as het.
		low_penetrance_genes -  In the autosomal_dominant inhertiance pattern gene symbols in this list will be treated as low penetrance and unaffected samples can have the variant.
		min_parental_gq_dn - Min parental genome quality score for de novo calculations
		min_parental_depth_dn - Min parental depth for de novo calculations
		min_parental_gq_upi - Min parental genome quality score for uniparental isodisomy calculations
		min_parental_depth_upi - Min parental depth for uniparental isodisomy calculations


		Returns:

		True - if the variant matches one of the inhertiance models in the models variable
		False - otherwise.
		"""

		if 'na' in models:

			return True

		if 'autosomal_dominant' in models:

			if self.matches_autosomal_dominant(lenient=lenient, low_penetrance_genes= low_penetrance_genes):

				return True

		if 'autosomal_reccessive' in models:

			if self.matches_autosomal_reccessive():

				return True

		if 'x_reccessive' in models:

			if self.matches_x_reccessive():

				return True

		if 'x_dominant' in models:

			if self.matches_x_dominant():

				return True

		if 'de_novo' in models:

			if self.matches_denovo(min_parental_gq=min_parental_gq_dn, min_parental_depth=min_parental_depth_dn, max_parental_alt_ref_ratio=max_parental_alt_ref_ratio_dn):

				return True

		if 'uniparental_isodisomy' in models:

			if self.matches_uniparental_isodisomy(min_parental_depth=min_parental_depth_upi, min_parental_gq=min_parental_gq_upi):

				return True

		if 'mitochrondrial' in models:

			if self.matches_mitochrondrial():

				return True

		if 'y_chrom' in models:

			if self.matches_y():

				return True

		if 'compound_het' in models:

			if self.matches_compound_het(compound_het_dict):

				return True

		return False

	def get_matching_inheritance_models(self, compound_het_dict, lenient=False, low_penetrance_genes={}, min_parental_gq_dn=30, min_parental_depth_dn=10, max_parental_alt_ref_ratio_dn=0.04, min_parental_gq_upi=30, min_parental_depth_upi=10):
		"""
		Return a list of all matching inheritance models.

		Input:

		models: list of models - possibilities are:
		autosomal_dominant
		autosomal_reccessive
		x_reccessive
		x_dominant
		de_novo
		uniparental_isodisomy
		mitochrondrial
		y_chrom

		compound_het_dict - Dictionary containing variants which are compound hets
		lenient - In the autosomal_dominant inhertiance pattern whether to allow affected other than the proband to be hom_alt as well as het.
		low_penetrance_genes -  In the autosomal_dominant inhertiance pattern gene symbols in this list will be treated as low penetrance and unaffected samples can have the variant.
		min_parental_gq_dn - Min parental genome quality score for de novo calculations
		min_parental_depth_dn - Min parental depth for de novo calculations
		min_parental_gq_upi - Min parental genome quality score for uniparental isodisomy calculations
		min_parental_depth_upi - Min parental depth for uniparental isodisomy calculations


		Returns:

		A list of all matching inheritance models.
		"""

		matching_models =[]

		if self.matches_autosomal_dominant(lenient=lenient, low_penetrance_genes= low_penetrance_genes):

			matching_models.append('autosomal_dominant')

		if self.matches_autosomal_reccessive():

			matching_models.append('autosomal_reccessive')

		if self.matches_x_reccessive():

			matching_models.append('x_reccessive')

		if self.matches_x_dominant():

			matching_models.append('x_dominant')

		if self.matches_denovo(min_parental_gq=min_parental_gq_dn, min_parental_depth=min_parental_depth_dn, max_parental_alt_ref_ratio=max_parental_alt_ref_ratio_dn):

			matching_models.append('de_novo')

		if self.matches_uniparental_isodisomy(min_parental_depth=min_parental_depth_upi, min_parental_gq=min_parental_gq_upi):

			matching_models.append('uniparental_isodisomy')

		if self.matches_mitochrondrial():

			matching_models.append('mitochrondrial')

		if self.matches_y():

			matching_models.append('y_chrom')

		if self.matches_compound_het(compound_het_dict):

			matching_models.append('compound_het')

		return matching_models



	def passes_gt_filter(self, family_member_id, min_dp=20, min_gq=30):
		"""
		Does the genotypes for a sample pass minimum depth and genotype quality requirements?

		Input:
		
		family_member_id: The family_member_id of the genotype the user wants to check
		min_dp: The minimum depth required to pass
		min_gq: The minimin genome quality required to pass.

		Returns:

		True - if the variant passes the filter
		False - otherwise.

		"""

		if self.get_depth(family_member_id) >= min_dp and self.get_genotype_quality(family_member_id) >= min_gq:

			return True

		return False


	def passes_filter(self):
		"""
		Does the variant pass the variant level filter?

		Input: self

		Returns:

		True - if the filter is PASS or None
		False - otherwise.

		"""

		if len(self.filter_status) == 1:

			if self.filter_status[0] == 'PASS' or self.filter_status[0] == None:

				return True


		return False

	def transcript_has_any(feature_key, value):
		"""
		Do any of the transcript annotations for feature_key qual the value?

		For example does the consequence for any transcript equal missense_variant?

		"""

		if self.transcript_annotations == None:

			return False

		for transcript in self.transcript_annotations:

			if transcript[feature_key] == value:

				return True

		return False


	
	def is_hom_alt_in_unaffected(self):
		"""
		Is the variant homozygous for the alt allele in unaffected samples?

		Input: self

		Returns:

		True - If the variant is homozygous alt in any unaffected sample
		False - Otherwise


		"""
		
		unaffected = self.family.get_unaffected_family_members()
		
		for family_member_id in unaffected:
			
			if self.is_hom_alt(family_member_id) == True:
				
				return True
			
		return False
	
	def is_on_autosome_or_xfemale(self):
		"""
		Is the variant on an autosome or on the x chromosome in a female sample.

		Applies to proband only.

		Input: self

		Returns:

			True - If the variant is on an autosome or on the x chromosome in a female
			False - otherwise


		"""
		
		proband = self.family.get_proband()
		
		if self.chrom not in ['X', 'Y', 'MT', 'M'] or (proband.sex == 2 and self.chrom =='X'):
		
			return True
		
		return False


	def add_transcript_annotations(self, transcript_annotations):
		"""
		Add a transcript annotation. This should be a list of dictionaries

		Input:
		
		transcript_annotations: A list of dictionaries containing transcript annotations.

		Returns:

		None

		"""

		assert isinstance(transcript_annotations, list)

		self.transcript_annotations = transcript_annotations

	def add_info_annotations(self, info_annotations):
		"""
		Add an info annotation. This should be a dictionary.

		Input:
		
		transcript_annotations: dictionary containing the info annotations.

		Returns:

		None

		"""

		assert isinstance(info_annotations, dict)

		self.info_annotations = info_annotations

	def get_numerical_transcript_annotation(self, annotation_key, zero_values=['.', '', None], agg_func='min'):
		"""
		Returns a numerical transcript annotation. If there are multiple values these can be aggregated using
		the add_func option.

		Input:
		
		annotation_key - A key to the numerical value in the self.transcript_annotations list of dicts
		zero_values - Values which should be changed to zero e.g. '.'
		agg_func - How to aggregate multiple values - possibilities = min, max, mean


		Returns:

		The numerical value as a float.
		"""

		annotations = []

		for transcript in self.transcript_annotations:

			annotation = transcript[annotation_key]

			if annotation in zero_values:

				annotations.append(0.0)

			elif '&' in annotation:

				annotation = annotation.split('&')

				for sub_annotation in annotation:

					if sub_annotation in zero_values:

						annotations.append(0.0)

					else:

						annotations.append(sub_annotation)


			else:

				annotations.append(annotation)

		annotations = [float(annotation) for annotation in annotations]

		if agg_func  == 'min':

			return min(annotations)

		elif agg_func == 'max':

			return max(annotations)

		elif agg_func == 'mean':

			return statistics.mean(annotations)

		else:

			raise ValueError('Invalid aggregation function supplied')


	def filter_on_numerical_transcript_annotation_gte(self,
												 annotation_key,
												 ad_het,
												 ad_hom_alt,
												 x_male,
												 x_female_het,
												 x_female_hom,
												 compound_het,
												 y,
												 mt,
												 zero_values=['.', '', None],
												 agg_func='min',
												 compound_het_dict = {}):

		"""
		Filter on a numerical value. Is the value is larger or equal to the provided value?

		Different values can be provided for different variant types. For example we can provide different 
		values for ad_het (autosomal heterozygous) and x_male (variants on x chromsome in men).

		Uses the genotype of whichever sample is set as the proband.

		Input:

			annotation_key - which annotation in the self.transcript_annotations list of dictionaries to get
			ad_het - the value to filter on for variants in which the proband is heterozygouse on an autosome.
			ad_hom_alt - the value to filter on for variants in which the proband is homozygous alt on an autosome.
			x_male - the value to filter on for variants on the X chromsome in which the proband is male.  
			x_female_het - the value to filter on for heterozygous variants on the X chromsome in which the proband is female. 
			x_female_hom - the value to filter on for homozygous variants on the X chromsome in which the proband is female.
			compound_het - the value to filter on if the variant is a compound het.
			y - the value to filter on if the variant is on the Y chromosome.
			mt - teh value to filter on if the variant is on the Mitochondrial chromosome.
			zero_values - list of values which should be considered as 0 in the annotation.
			agg_func - how to aggregate multiple values - min, max, mean
			compound_het_dict - A dictionary of variants which are compound hets.

		Returns:

			True - if the variant has a value above the supplied value
			False - otherwise.

		"""


		annotation = self.get_numerical_transcript_annotation(annotation_key, zero_values, agg_func)
		proband = self.family.get_proband()
		proband_id = proband.get_id()


		if self.variant_id in compound_het_dict:

			if annotation >= compound_het:

					return True

			else:

				return False


		if self.chrom not in ['X', 'Y', 'MT', 'M'] and self.is_het(proband_id):

			if annotation >= ad_het:

				return True

			else:

				return False

		elif self.chrom not in ['X', 'Y', 'MT', 'M'] and self.is_hom_alt(proband_id):

			if annotation >= ad_hom_alt:

				return True

			else:

				return False

		elif self.chrom == 'X' and proband.sex == 1 and self.has_alt(proband_id):

			if annotation >= x_male:

				return True

			else:

				return False	

		elif self.chrom == 'X' and proband.sex == 2 and self.is_het(proband_id):

			if annotation >= x_female_het:

				return True

			else:

				return False	

		elif self.chrom == 'X' and proband.sex == 2 and self.is_hom_alt(proband_id):

			if annotation >= x_female_hom:

				return True

			else:

				return False

		elif self.chrom =='Y':

			if annotation >= y:

				return True

			else:

				return False

		elif self.chrom =='MT' or self.chrom =='M':

			if annotation >= mt:

				return True

			else:

				return False

		else:

			raise ValueError('Variant does not have a a recognised workflow - is the proband homozygous reference?')



	def filter_on_numerical_transcript_annotation_lte(self,
												 annotation_key,
												 ad_het,
												 ad_hom_alt,
												 x_male,
												 x_female_het,
												 x_female_hom,
												 compound_het,
												 y,
												 mt,
												 zero_values=['.', '', None],
												 agg_func='min',
												 compound_het_dict = {}):
		"""
		Filter on a numerical value. Is the value is less than or equal to the provided value?

		Different values can be provided for different variant types. For example we can provide different 
		values for ad_het (autosomal heterozygous) and x_male (variants on x chromsome in men).

		Uses the genotype of whichever sample is set as the proband.

		Input:

			annotation_key - which annotation in the self.transcript_annotations list of dictionaries to get
			ad_het - the value to filter on for variants in which the proband is heterozygouse on an autosome.
			ad_hom_alt - the value to filter on for variants in which the proband is homozygous alt on an autosome.
			x_male - the value to filter on for variants on the X chromsome in which the proband is male.  
			x_female_het - the value to filter on for heterozygous variants on the X chromsome in which the proband is female. 
			x_female_hom - the value to filter on for homozygous variants on the X chromsome in which the proband is female.
			compound_het - the value to filter on if the variant is a compound het.
			y - the value to filter on if the variant is on the Y chromosome.
			mt - teh value to filter on if the variant is on the Mitochondrial chromosome.
			zero_values - list of values which should be considered as 0 in the annotation.
			agg_func - how to aggregate multiple values - min, max, mean
			compound_het_dict - A dictionary of variants which are compound hets.

		Returns:

			True - if the variant has a value below the supplied value
			False - otherwise.

		"""


		annotation = self.get_numerical_transcript_annotation(annotation_key, zero_values, agg_func)
		proband = self.family.get_proband()
		proband_id = proband.get_id()

		if self.variant_id in compound_het_dict:

			if annotation <= compound_het:

				return True

			else:

				return False

		elif self.chrom not in ['X', 'Y', 'MT', 'M'] and self.is_het(proband_id):

			if annotation <= ad_het:

				return True

			else:

				return False

		elif self.chrom not in ['X', 'Y', 'MT', 'M'] and self.is_hom_alt(proband_id):

			if annotation <= ad_hom_alt:

				return True

			else:

				return False

		elif self.chrom == 'X' and proband.sex == 1 and self.has_alt(proband_id):

			if annotation <= x_male:

				return True

			else:

				return False	

		elif self.chrom == 'X' and proband.sex == 2 and self.is_het(proband_id):

			if annotation <= x_female_het:

				return True

			else:

				return False	

		elif self.chrom == 'X' and proband.sex == 2 and self.is_hom_alt(proband_id):

			if annotation <= x_female_hom:

				return True

			else:

				return False

		elif self.chrom =='Y':

			if annotation <= y:

				return True

			else:

				return False

		elif self.chrom =='MT' or self.chrom == 'M':

			if annotation <= mt:

				return True

			else:

				return False

		else:

			raise ValueError('Variant does not have a a recognised workflow - is the proband homozygous reference?')


	def matches_paternal_uniparental_ambiguous(self,  min_parental_gq=30, min_parental_depth=10):
		"""
		A type of mendelian error symptomatic of a UPD event.

		Where the dad has allele AA and mum BB and child has AA

		This could happen from either a heterodisomic or isodisomic event

		"""

		proband = self.family.get_proband()
		
		mum = proband.get_mum()
		dad = proband.get_dad()

		# Must have both parents to calculate
		if mum == None or dad == None:
				
			return False

		mum_id = mum.get_id()
		dad_id = dad.get_id()
		proband_id = proband.get_id()

		mum_gq = self.get_genotype_quality(mum_id)
		mum_dp = self.get_depth(mum_id)
		dad_gq = self.get_genotype_quality(dad_id)
		dad_dp = self.get_depth(dad_id)

		# dad must be homozygous and mum must be homozygous for other allele and child homozygouse for same allele as dad


		if (self.is_hom_ref(dad_id) and
			self.is_hom_alt(mum_id) and 
			self.is_hom_ref(proband_id) and
			mum_gq >= min_parental_gq and
			mum_dp >= min_parental_depth and
			dad_gq >= min_parental_gq and
			dad_dp >= min_parental_depth):
			return True

		if (self.is_hom_alt(dad_id) and
			self.is_hom_ref(mum_id) and 
			self.is_hom_alt(proband_id) and
			mum_gq >= min_parental_gq and
			mum_dp >= min_parental_depth and
			dad_gq >= min_parental_gq and
			dad_dp >= min_parental_depth):
			return True 

		return False


	def matches_maternal_uniparental_ambiguous(self,  min_parental_gq=30, min_parental_depth=10):
		"""
		A type of mendelian error symptomatic of a UPD event.

		Where the dad has allele AA and mum BB and child has AA

		This could happen from either a heterodisomic or isodisomic event

		"""

		proband = self.family.get_proband()
		
		mum = proband.get_mum()
		dad = proband.get_dad()

		# Must have both parents to calculate
		if mum == None or dad == None:
				
			return False

		mum_id = mum.get_id()
		dad_id = dad.get_id()
		proband_id = proband.get_id()

		mum_gq = self.get_genotype_quality(mum_id)
		mum_dp = self.get_depth(mum_id)
		dad_gq = self.get_genotype_quality(dad_id)
		dad_dp = self.get_depth(dad_id)

		# dad must be homozygous and mum must be homozygous for other allele and child homozygouse for same allele as dad


		if (self.is_hom_ref(mum_id) and
			self.is_hom_alt(dad_id) and 
			self.is_hom_ref(proband_id) and
			mum_gq >= min_parental_gq and
			mum_dp >= min_parental_depth and
			dad_gq >= min_parental_gq and
			dad_dp >= min_parental_depth):
			return True

		if (self.is_hom_alt(mum_id) and
			self.is_hom_ref(dad_id) and 
			self.is_hom_alt(proband_id) and
			mum_gq >= min_parental_gq and
			mum_dp >= min_parental_depth and
			dad_gq >= min_parental_gq and
			dad_dp >= min_parental_depth):
			return True 

		return False

	def matches_paternal_uniparental_isodisomy(self,  min_parental_gq=30, min_parental_depth=10):

		proband = self.family.get_proband()
		
		mum = proband.get_mum()
		dad = proband.get_dad()

		# Must have both parents to calculate
		if mum == None or dad == None:
				
			return False

		mum_id = mum.get_id()
		dad_id = dad.get_id()
		proband_id = proband.get_id()

		mum_gq = self.get_genotype_quality(mum_id)
		mum_dp = self.get_depth(mum_id)
		dad_gq = self.get_genotype_quality(dad_id)
		dad_dp = self.get_depth(dad_id)

		# mum homozygous ref dad het and kid homozygous for alt
		if (self.is_hom_ref(mum_id) and 
			self.is_het(dad_id) and 
			self.is_hom_alt(proband_id) and
			mum_gq >= min_parental_gq and
			mum_dp >= min_parental_depth and
			dad_gq >= min_parental_gq and
			dad_dp >= min_parental_depth):

			return True

		# mum homozygous alt dad het and kid homozygous for ref
		if (self.is_hom_alt(mum_id) and 
			self.is_het(dad_id) and 
			self.is_hom_ref(proband_id) and
			mum_gq >= min_parental_gq and
			mum_dp >= min_parental_depth and
			dad_gq >= min_parental_gq and
			dad_dp >= min_parental_depth):

			return True

		return False


	def matches_maternal_uniparental_isodisomy(self,  min_parental_gq=30, min_parental_depth=10):

		proband = self.family.get_proband()
		
		mum = proband.get_mum()
		dad = proband.get_dad()

		# Must have both parents to calculate
		if mum == None or dad == None:
				
			return False

		mum_id = mum.get_id()
		dad_id = dad.get_id()
		proband_id = proband.get_id()

		mum_gq = self.get_genotype_quality(mum_id)
		mum_dp = self.get_depth(mum_id)
		dad_gq = self.get_genotype_quality(dad_id)
		dad_dp = self.get_depth(dad_id)

		# mum homozygous ref dad het and kid homozygous for alt
		if (self.is_hom_ref(dad_id) and 
			self.is_het(mum_id) and 
			self.is_hom_alt(proband_id) and
			mum_gq >= min_parental_gq and
			mum_dp >= min_parental_depth and
			dad_gq >= min_parental_gq and
			dad_dp >= min_parental_depth):

			return True

		# mum homozygous alt dad het and kid homozygous for ref
		if (self.is_hom_alt(dad_id) and 
			self.is_het(mum_id) and 
			self.is_hom_ref(proband_id) and
			mum_gq >= min_parental_gq and
			mum_dp >= min_parental_depth and
			dad_gq >= min_parental_gq and
			dad_dp >= min_parental_depth):

			return True

		return False


	def alleles_identical_to_dad(self, min_parental_gq=30, min_parental_depth=10):
		"""
		Does the proband have identical alleles to dad?

		"""

		proband = self.family.get_proband()
		
		mum = proband.get_mum()
		dad = proband.get_dad()

		# Must have both parents to calculate
		if mum == None or dad == None:
				
			return False

		mum_id = mum.get_id()
		dad_id = dad.get_id()
		proband_id = proband.get_id()

		if self.is_missing(proband_id) or self.is_missing(dad_id):

			return False

		dad_gq = self.get_genotype_quality(dad_id)
		dad_dp = self.get_depth(dad_id)

		gt_proband = self.genotypes[proband_id]['genotype']
		gt_dad = self.genotypes[dad_id]['genotype']

		if (set(gt_proband) == set(gt_dad) and 
			dad_gq >= min_parental_gq and
			dad_dp >= min_parental_depth):

			return True

		return False

	def alleles_identical_to_mum(self, min_parental_gq=30, min_parental_depth=10):
		"""
		Does the proband have identical alleles to mum?

		"""

		proband = self.family.get_proband()
		
		mum = proband.get_mum()
		dad = proband.get_dad()

		# Must have both parents to calculate
		if mum == None or dad == None:
				
			return False

		mum_id = mum.get_id()
		dad_id = dad.get_id()
		proband_id = proband.get_id()

		if self.is_missing(proband_id) or self.is_missing(mum_id):

			return False

		mum_gq = self.get_genotype_quality(mum_id)
		mum_dp = self.get_depth(mum_id)

		gt_proband = self.genotypes[proband_id]['genotype']
		gt_mum = self.genotypes[mum_id]['genotype']

		if (set(gt_proband) == set(gt_mum) and 
			mum_gq >= min_parental_gq and
			mum_dp >= min_parental_depth):

			return True

		return False

	def is_snp(self):
		"""
		Is the variant a simple snp with a ref and alt length of 1?
		"""

		if len(self.ref) == 1 and len(self.alt) == 1:

			return True

		return False


	def all_samples_pass_genotype_quality(self , min_dp=10, min_gq=20):
		"""
		Check all sample genotypes are not missing and have depths and gqs above the minimum.
		"""

		for sample_id in self.family.get_all_family_member_ids():

			gq = self.get_genotype_quality(sample_id)
			dp = self.get_depth(sample_id)

			if dp < min_dp:

				return False

			if gq < min_gq:

				return False

			if self.is_missing(sample_id) == True:

				return False

		return True

	def get_quality(self):
		"""	
		Return the quality
		"""

		return self.quality


	def is_biparental_inheritance(self, min_parental_gq=30, min_parental_depth=10):
		"""
		Is the site one in which the proband inherits a different allele from each parent?

		Or at least a site where this is provable who transmits what!

		Can be used to rule out UPD across a region

		"""

		proband = self.family.get_proband()
		
		mum = proband.get_mum()
		dad = proband.get_dad()

		# Must have both parents to calculate
		if mum == None or dad == None:
				
			return False

		mum_id = mum.get_id()
		dad_id = dad.get_id()
		proband_id = proband.get_id()

		mum_gq = self.get_genotype_quality(mum_id)
		mum_dp = self.get_depth(mum_id)

		dad_gq = self.get_genotype_quality(dad_id)
		dad_dp = self.get_depth(dad_id)

		if self.is_missing(proband_id) or self.is_missing(mum_id):

			return False

		# dad is hom for alt mum is hom for ref and kid is het

		if (self.is_hom_alt(dad_id) and 
			self.is_hom_ref(mum_id) and 
			self.is_het(proband_id) and
			mum_gq >= min_parental_gq and
			mum_dp >= min_parental_depth and
			dad_gq >= min_parental_gq and
			mum_dp >= min_parental_depth):

			return True

		# dad is hom for ref mum is hom for alt and kid is het
		if (self.is_hom_ref(dad_id) and 
			self.is_hom_alt(mum_id) and 
			self.is_het(proband_id) and
			mum_gq >= min_parental_gq and
			mum_dp >= min_parental_depth and
			dad_gq >= min_parental_gq and
			mum_dp >= min_parental_depth):

			return True	

		return False


import setuptools

with open('readme.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='pyvariantfilter',
    version='1.0.3',
    author='Joseph Halstead',
    author_email='josephhalstead89@gmail.com',
    description='Python package for filtering variants.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/josephhalstead/pyvariantfilter',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
    install_requires=[
   'pysam>=0.15.2',
   'pandas>=0.23.4'
],
)
